//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//#
//#			Apple Macintosh Developer Technical Support
//#
//#			Standard File Sample Application
//#
//#			StdFile
//#
//#			StdFile.p			-			Pascal Source
//#
//#			Copyright  1989 Apple Computer, Inc.
//#			All rights reserved.
//#
//#		Versions:		
//#					1.00					04/89
//#					2.00					05/90
//#					2.01					06/92
//#
//#		Components:
//#					StdFile.p				April 1, 1989
//#					StdFile.h				April 1, 1988
//#					StdFile.r				April 1, 1988
//#					PStdFile.make			April 1, 1988
//#
//# ----------
//# OBJECTIVES
//# ----------
//# This program attempts to demonstrate the following techniques:
//#
//# - Normal use of SFGetFile and SFPutFile
//# - Normal use of SFPGetFile and SFPPutFile.	This includes use of Custom
//#		Dialogs with handling of extra items through the implementation of a
//#		Standard File Dialog Hook.
//#		- First time initialization
//#		- Extra Simple buttons (Disory/ThisDir/Options)
//#		- Radio buttons (file format, types of files to show)
//#		- Aliasing click on some buttons to clicks on other buttons
//#		- Regenerating the list of files displayed
//# - Creating a full pathname from the reply record (using Working Directories
//#		or DirID)
//# - Selecting a directory (ala MPW's "GetFileName -d")
//# - Simple file filter (checks file type)
//# - Complex file filter (looking inside the file)
//# - Adding and deleting and extra List Manager lists.		This is shown for both
//#		SFGetFile and SFPutFile.
//# - Select multiple files by adding a second list to the Dialog Box.
//# - Setting the starting Directory/Volume
//# - Describe pending update event clogging
//# - Ask the user to select a file, and then remember where it is the next
//#		time the application is launched.
//#
//# --------------------------
//# Standard File Generalities
//# --------------------------
//#
//# The prompt string of SFGetFile is ignored in all calls.		As noted on page
//# I-523 of Inside Mac, it is there for historical reasons only.	 The
//# Standard File package is built around a modal dialog window.	Standard
//# File repeatedly calls ModalDialog waiting for user events.	It is
//# possible for an application to hook into the dialog event loop by using
//# the following methods.
//#
//# File Filter
//# -----------
//# The file filter is called to give the application opportunity to filter
//# out files that is wishes not to be shown in the dialog.		This file filter
//# is called before displaying the dialog, which is also before the Dialog
//# Hook or Dialog Filter are called.	 If this function returns TRUE, this
//# means the file should not be shown in the list.		In other words the file
//# is to be filtered from the user.
//#
//# Dialog Hook
//# -----------
//# The dialog hook is called after Standard File calls ModalDialog.	The
//# hook is a function that requires that an item number be passed back from
//# it.		Normally, this is the same item number that was passed to us, but
//# not necessarily.	For instance, we could return values that cause the
//# file names to be re-drawn or have the whole event ignored.	Before the
//# dialog is drawn, our dialog hook gets called with a -1.		This gives us
//# the opportunity to change things like Button titles, etc.	 It's in
//# important to note that Standard File seems to ignore the item returned by
//# the dialog hook when it is called the first time.	 The dialog hook will
//# also be passed a fake item of 100, which is used to mean a nullEvent.
//# This can be used as an idle event for the dialog hook.
//#
//# Dialog Filter
//# -------------
//# This is used in the SFPGetFile and SFPPutFile routines.		It is for the
//# advanced versions of calling Standard File for programmers that desire a
//# custom dialog window.	 The dialog filter is the same filter procedure
//# used by the Dialog Manager.		In the call to ModalDialog, you can supply a
//# filter.		The dialog filter used by Standard File is passed to
//# ModalDialog.	This filter is called by the Dialog Manager after it calls
//# GetNextEvent.	 This gives the filter the first opportunity to handle the
//# event.	If the dialog filter returns TRUE, this tells the Dialog Manager
//# that the filter has handled the event.	You will use the dialog filter to
//# handle items you've added to your custom dialog.
//#
//#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#include <stddef.h>
#include <types.h>
#include <limits.h>
#include "Builtins.h"
#include "PStrings.h"
#include "PSets.h"

#include "Types.h"
#include <QuickDraw.h>
#include "ToolUtils.h"
#include "Events.h"
#include "Controls.h"
#include "Windows.h"
#include "Dialogs.h"
#include "Menus.h"
#include "Devices.h"
#include "SegLoad.h"
#include "Files.h"
#include "Traps.h"
#include "Fonts.h"
#include "Resources.h"
#include "Memory.h"
#include "Packages.h"
#include "Lists.h"
#include "DiskInit.h"
#include "Script.h"
#include "LowMem.h"
#include <CursorCtl.h>
#include "Errors.h"
#include "FixMath.h"
#include "GestaltEqu.h"
#include "StandardFile.h"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//	This record is enough to specify a file before rebooting.
//	File names were limited to 63 chars with MFS, but HFS limits names to 31 chars

struct FileSpec {
	short vRefNum;
	long parID;
	Str31 name;
};
typedef struct FileSpec FileSpec;

typedef FileSpec FileSpecList[1000];
typedef FileSpecList *FileSpecPtr;
typedef FileSpecPtr *FileSpecHdl;

//This record is enough to specify a file after rebooting.	This record is kept
//as a resource when the user selects a file.

struct FileDesc {
	Str27 volName;		//	volume names are limited to 27 chars
	long parID;
	Str31 name;			//	file names are limited to 31 chars
};
typedef struct FileDesc FileDesc;
typedef FileDesc *FileDescPtr, **FileDescHdl;

//This record is used for the unsigned long integers which Pascal really doesn't handle

union TwoIntsMakesALong {							//build a signed LONGINT...
													//...from an unsigned INTEGER
	long long_1;
	short ints[2];
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//gMac is used to hold the result of a SysEnvirons call. This makes it
//convenient for any routine to check the environment. It is considered
//global information, anyway.

SysEnvRec gMac;										//set up by Initialize

//gInBackground is maintained by our osEvent handling routines. Any part of
//the program can check it to find out if it is currently in the background.

Boolean gInBackground;								//maintained by Initialize and EventLoop

//This determines if we can call WaitNextEvent, which is always present for any
//System 6.0x user but only possible for earilier systems running MultiFinder.

Boolean gHasWaitNextEvent;							//set by Initialize

//These next three globals are used to remember the application's location.
//This is the volume and the folder that hold the application at launch time.

Str255 gAppName;									//application's name
long gAppParID;										//dirID of application's folder
short gAppVRefNum;									//volume containing application

//Sometimes it is necessary to have a global SFReply record.

SFReply gReply;										//used in some SF samples

//This is a flag used to signal a normal file format or not used in a couple examples.

Boolean gNormal;

//This is a handle to an array of my FileSpecs.		It is accessed by a few routines, so
//it needed to be global.

FileSpecHdl gMultiFiles;							//array of files to open

//This is a handle to a List Manager list.	It is accessed by a few routines, so
//it needed to be global.

ListHandle gListHandle;								//list use for multiple get files

//This is the selected dirID that the user selected while in SFGet.

long gMyCurDir;										//for SetDirectory sample

//This is a string set by a few of the examples used to determine which file format
//the user wanted to save the file as.

Str255 gPutFormat;									//string of file format

long gLsBkUp;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//InitGraf is always implemented (trap $A86E). If the trap table is big
//enough, trap $AA6E will always point to either Unimplemented or some other
//trap, but will never be the same as InitGraf. Thus, you can check the size
//of the trap table by asking if the address of trap $A86E is the same as $AA6E.

short NumToolboxTraps(void)
{
	short value_NumToolboxTraps;

	if (NGetTrapAddress(_InitGraf, ToolTrap) == NGetTrapAddress(0xAA6E, 
																ToolTrap)) 
	value_NumToolboxTraps = 0x200;
	else 
	value_NumToolboxTraps = 0x400;
	return value_NumToolboxTraps;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//Determines the type of a trap based on its trap number. If bit 11 is clear,
//then it is an OS trap. Otherwise, it is a Toolbox trap.

TrapType GetTrapType(short theTrap)
{
	// OS traps start with A0, Tool with A8 or AA. 
	TrapType value_GetTrapType;

	if (BAnd(theTrap, 0x0800) == 0)					//per Darin
	value_GetTrapType = OSTrap;
	else 
	value_GetTrapType = ToolTrap;
	return value_GetTrapType;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Boolean TrapExists(short theTrap)
{
	// Check to see if a given trap is implemented. This is only used by the
	// Initialize routine in this program, so we put it in the Initialize segment. 
	
	TrapType theTrapType;

	theTrapType = GetTrapType(theTrap);
	if (theTrapType == ToolTrap) {
	theTrap = BAnd(theTrap, 0x07FF);
	if (theTrap >= NumToolboxTraps()) 
		theTrap = _Unimplemented;
	}
	
	return NGetTrapAddress(_Unimplemented, ToolTrap) != NGetTrapAddress(theTrap, 
																		theTrapType);
} // TrapExists 

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

short GetAUXVersion(void)
{
	//The glue code for Gestalt will return appropriate values if the
	//system doesn't support the trap.	If an error is returned from Gestalt,
	//then either we weren't running on A/UX, or the Gestalt trap
	//was unable to fulfill the request so we revert to using low memory
	//globals instead.	This will work for A/UX version 1.x.	All other errors
	//are ignored (implies A/UX not present).
	
	//NOTE: We right shift AUXversion by 8 bits to get major version number
	//contained in the high byte, since we're not concerned with the minor
	//version.	Anything better than 2.0 is good enough for me.
	
	const char	kAUXbit = 9;							//bit set in LMGetHWCfgFlags for A/UX
	const char	kShiftHighByte = 8;						//shift 8 bits to the right
	
	long AUXversion;
	short err;

	err = Gestalt(gestaltAUXVersion, &AUXversion);
	
	if (err != noErr) {
	if (BTst(LMGetHWCfgFlags(), kAUXbit)) 
		AUXversion = kDefaultAUXVersion;				//A/UX is running, default version
	else 
		AUXversion = 0;								//less than 1 means no version
	}
	return BSR(AUXversion, kShiftHighByte);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//The following set of routines are used to access a couple of low memory
//globals that are necessary when extending Standard File.	One example is
//trying to get the current directory while in a file filter.	 These routines
//were used to bottleneck all the low memory usage.		If the system one day
//supports them with a trap call, then we can easily update these routines.

long GetSFCurDir(void)
{
	return LMGetCurDirStore();
}

void SetSFCurDir(long dirID)
{
	LMSetCurDirStore(dirID);
}

short GetSFVRefNum(void)
{
	return -LMGetSFSaveDisk();
}

void SetSFVRefNum(short vRefNum)
{
	LMSetSFSaveDisk(-vRefNum);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

long GetKFreeSpace(short vRefNum)
{
	//Return the amount of free space on the volume in KBytes. This builds a
	// LONGINT based on an unsigned INTEGER, which looks like a negitive value to
	// Pascal.	-1 is return as the size if there is an error.
	
	HParamBlockRec pb;
	TwoIntsMakesALong convert;
	OSErr err;
	long value_GetKFreeSpace;

													//set up the block for PBHGetVInfo
	pb.ioParam.ioNamePtr = NULL;						//we don't care about the name
	pb.ioParam.ioVRefNum = vRefNum;
	pb.volumeParam.ioVolIndex = 0;					//use ioVRefNum only
	err = PBHGetVInfo(&pb, false);
	
	if (err == noErr) {
	convert.ints[0] = 0;
	convert.ints[1] = pb.volumeParam.ioVFrBlk;
	value_GetKFreeSpace = (convert.long_1 * pb.volumeParam.ioVAlBlkSiz) / 1024;
	} else											//we couldn't get free space size!
	value_GetKFreeSpace = -1;
	return value_GetKFreeSpace;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Boolean FailLowMemory(long memRequest)
{
	//Call PurgeSpace and see if the requested amount of memory exists in the
	//heap including a minimal amount of heap space.	Also, if my grow zones
	//reserve has been release, thats considered a failure.	I dont perform
	//any purging here.	 The Memory Manager will do this if it needs the space.
	//This routine can be called with a memRequest = 0.	 This checks if the heap
	//space is getting critical.
	
	long total;
	long contig;

	PurgeSpace(&total, &contig);
	return total < memRequest + kMinSpace;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

short GetWTitleHeight(short variant)
{
	//Try and determine the windows title bar height.	This isnt easy,
	//especially if the window is invisible.	We all know how the standard Apple
	//System WDEF works, at least at this date we do.		I assume that method as
	//shown below.	I could do what MacApp does with the Function CallWDefProc.
	//If the user has installed a replacement for the System WDEF, then its
	//their problem to deal with.		My method will work as long as Apple doesnt
	//change how the WDEF in System 6.0 calculates the title bar height.	This
	//will allow my application to work on international Macs that have a larger
	//system font than Chicago.	 I check the windows variant code for one that
	//includes a title bar.	 I will have to change this routine to adjust for
	//the modal-moveable window type, which hasnt been defined yet.
	
	//In this routine, I violate my rule about not using the GetPort, SetPort,
	//SetPort sequence mentioned at the start of the file. Mostly, I do this
	//because its not all that apparent that a routine called GetWTitleHeight
	//will change the port, so I make sure that it doesnt.
	
	FontInfo info;
	GrafPtr curGraf;
	GrafPtr wMgrPort;
	short wTitleHeight;
	short value_GetWTitleHeight;

	if ((1<<variant) & ((1<<documentProc)|(1<<noGrowDocProc)|(1<<zoomDocProc)|(1<<zoomNoGrow)|(1<<rDocProc))) {
	GetPort(&curGraf);
	GetWMgrPort(&wMgrPort);							//I need to know the font...
	SetPort(wMgrPort);								//info in the Systems port
	GetFontInfo(&info);
	SetPort(curGraf);								//restore current port
	wTitleHeight = info.ascent + info.descent + info.leading + 2;
	if (wTitleHeight < 19) 
		wTitleHeight = 19;							//the title is always at least 19
	value_GetWTitleHeight = wTitleHeight;
	} else 
	value_GetWTitleHeight = 0;						//other window types have no title
	return value_GetWTitleHeight;
} //GetWTitleHeight

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void CenterWindowRect(short variant, RectPtr theRect)
{
	//Given a windows portRect, this routine will center the rectangle on the
	//main device within the desktop region.	This excludes the menu bar area.
	//It follows the Apple Human Interface Guidelines for where to place a
	//window centered on the screen. That is, 1/3th of the desktop shows above
	//the window and 2/3ths below it.		It returns a new rectangle set to the
	//centered position.	The top and left of this rectangle can be used with
	//MoveWindow.		This routine only considers the main device.	CenterWindowRect
	//could take a GDevice as a parameter to center the VAR rect with.	This
	//would also allow Alerts or dialogs that are closely associated with a
	//document window to be centered relative to the monitor that contains that
	//document.	 This is also one of the Human Interface Guidelines.	MacApp 2.0
	//has a utility CalcScreenRect that you can borrow from to include this.
	
	//WARNING: This routine may move or purge memory.
	
	Point rectSize;
	short wTitleHeight;

	wTitleHeight = GetWTitleHeight(variant);			//get title height
													//get size of rect
	SetPt(&rectSize, theRect->right, theRect->bottom + wTitleHeight);		//include it in size
	SubPt(theRect->topLeft, &rectSize);
													//1/3th below menubar
	theRect->top = (((qd.screenBits.bounds.bottom - qd.screenBits.bounds.top) - LMGetMBarHeight()) - rectSize.v) / 3 + LMGetMBarHeight();
	theRect->left = ((qd.screenBits.bounds.right - qd.screenBits.bounds.left) - rectSize.h) / 2;	//centered horz
													//return adjusted rect
	SetPt(&theRect->botRight, theRect->left, theRect->top);
	AddPt(rectSize, &theRect->botRight);
	theRect->top += wTitleHeight;						//remove title height from rect
} //CenterWindowRect

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

short CenteredAlert(short alertID)
{
	//Given an Alert ID, this routine will center the alerts rectangle before
	//showing it on the main screen.	This follows the Apple Human Interface
	//Guidelines for where to place a centered window on the screen.	If the
	//Alert is closely associated with another window, considerations should be
	//given to what the window is located.	If this other window is on a screen
	//other then the main monitor, then it would be best to center the Alert on
	//that other monitor.
	
	AlertTHndl alertHandle;
	Rect alertRect;
	short itemHit;

	alertHandle = (AlertTHndl)Get1Resource('ALRT', alertID);
	if (alertHandle != NULL) {
	HNoPurge((Handle)alertHandle);
	alertRect = (*alertHandle)->boundsRect;
	CenterWindowRect(dBoxProc, &alertRect);
	(*alertHandle)->boundsRect = alertRect;
	HPurge((Handle)alertHandle);
	}
	return Alert(alertID, NULL);
} //CenteredAlert

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void EmergencyExit(short message)
{
	//Display an alert that tells the user an error occurred, then exit the
	//program.	This routine is used as an ultimate bail-out for serious errors
	//that prohibit the continuation of the application.	Errors that do not
	//require the termination of the application are handled with AlertUser.
	//Error checking and reporting has a place even in the simplest application.
	
	//BUG NOTE: GetIndString will return a bogus String if the index is not
	//positive.
	
	Str255 msg1;
	short theItem;

	SetCursor(&qd.arrow);
	if (message > 0) 
	GetIndString(msg1, rErrStrings, message);
	else 
	PStr_Set(msg1, "\p");							//in case there is no message
	ParamText(msg1, "\p", "\p", "\p");
	theItem = CenteredAlert(rExitAlert);
	ExitToShell();									//we're out of here
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void SetRadioButton(DialogPtr dlgPtr, short itemNo, short state)
{
	//Handy routine for setting the value of a radio button. Given a dialog
	// pointer, and item number, and a state, this routine will take care of the
	// rest.
	
	short iKind;
	Handle iHandle;
	Rect iRect;

	GetDialogItem(dlgPtr, itemNo, &iKind, &iHandle, &iRect);
	SetControlValue((ControlHandle)iHandle, state);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void PositionTwoRects(RectPtr outerRect, RectPtr innerRect, Fixed horzRatio, Fixed vertRatio)
{
	//Given two rectangles, this routine positions the second within the first one
	// so that the it maintains the spacing specified by the horzRatio and vertRatio
	// parameters. In other words, to center an inner rectangle hoizontally, but
	// have its center be 1/3 from the top of the outer rectangle, call this
	// routine with horzRatio = FixRatio (1, 2), vertRatio = FixRatio (1, 3).		We use
	// Fixed rather than floating point because Fixed has plenty of accuracy and is very fast. 
	
	short outerRectHeight;
	short outerRectWidth;
	short innerRectHeight;
	short innerRectWidth;
	short yLocation;
	short xLocation;

	outerRectHeight = outerRect->bottom - outerRect->top;
	outerRectWidth = outerRect->right - outerRect->left;
	
	innerRectHeight = innerRect->bottom - innerRect->top;
	innerRectWidth = innerRect->right - innerRect->left;
	
	yLocation = Fix2Long(FixMul(Long2Fix(outerRectHeight - innerRectHeight), 
								vertRatio)) + outerRect->top;
	xLocation = Fix2Long(FixMul(Long2Fix(outerRectWidth - innerRectWidth), 
								horzRatio)) + outerRect->left;
	innerRect->top = yLocation;
	innerRect->left = xLocation;
	innerRect->bottom = yLocation + innerRectHeight;
	innerRect->right = xLocation + innerRectWidth;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void OutlineButton(ControlHandle button)
{
	//Given any control handle, this will draw an outline around it.	This is
	//used for the default button of a window.	The extra nice feature here is
	//that Ill erase the outline for buttons that are inactive.	Seems like
	//there should be a Toolbox call for getting a controls hilite state.
	//Since there isnt, I have to look into the control record myself.	 This
	//should be called for update and activate events.
	
	//The method for determining the oval diameters for the roundrect is a
	//little different than that recommended by Inside Mac.	 IM I-407 suggests
	//that you use a hardcoded (16,16) for the diameters. However, this only
	//looks good for small roundrects. For larger ones, the outline doesnt
	//follow the inner roundrect because the CDEF for simply buttons doesnt
	//use (16,16). Instead, it uses half the height of the button as the
	//diameter. By using this formula, too, our outlines look better.
	
	//WARNING: This will set the current port to the controls window.
	
	Rect theRect;
	PenState curPen;
	short buttonOval;

	if (button != NULL) {
	SetPort((*button)->contrlOwner);
	GetPenState(&curPen);
	PenNormal();
	theRect = (*button)->contrlRect;
	InsetRect(&theRect, kButtonFrameInset, kButtonFrameInset);
	buttonOval = (theRect.bottom - theRect.top) / 2;
	if ((*button)->contrlHilite == kCntlActivate) 
		PenPat(&qd.black);
	else 
		PenPat(&qd.gray);
	PenSize(kButtonFrameSize, kButtonFrameSize);
	FrameRoundRect(&theRect, buttonOval, buttonOval);
	SetPenState(&curPen);
	}
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DoAbout(void)
{
	StringHandle apNameHndl;
	VersRecHndl curVersion;
	Str255 apName;
	Str255 verNum;
	short itemHit;

	curVersion = (VersRecHndl)Get1Resource('vers', 1);
	if (curVersion != NULL) {
	PStr_Set(verNum, ((StringPtr)((long)&(*curVersion)->shortVersion + (*curVersion)->shortVersion[0] + 1)));	 // get long version string 
	} else 
	PStr_Set(verNum, "\p????");						//at least initialize it
	
	ParamText(gAppName, verNum, "\p", "\p");
	itemHit = CenteredAlert(rAboutAlert);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void AlertUser(short error, short messageID)
{
	//Display an alert to inform the user of an error.	MessageID acts as an
	//index into a Str# resource of error messages. This will attempt to
	//replace the error number with a String if the sound is a Sound Manager
	//error.
	
	//BUG NOTE: GetIndString will return a bogus String if the index is not
	//positive.
	
	Str255 msg1;
	Str255 msg2;
	short theItem;

	if (messageID > 0) 
	GetIndString(msg1, rErrStrings, messageID);
	else 
	PStr_Set(msg1, "\p");							//in case there is no message
	if (error != noErr) 
	NumToString(error, msg2);
	else 
	PStr_Set(msg2, "\p");
	ParamText(msg1, msg2, "\p", "\p");
	theItem = CenteredAlert(rUserAlert);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	Don't need this.  Just call GetProcessInformation and use the 
//	app's FSSpec for the needed info, since the default directory,
//	for System 7 or later, will be the application's directory.

OSErr MyHGetVol(StringPtr volName, SInt16 *vRefNum, IOCommandID dirID)
{
	//BUG NOTE: The high level call, HGetVol, should do this dirty work for us.
	//Unfortunately it is returning the ioVRefNum which Inside Mac warns may be
	//a working directory.	The field ioWDVRefNum will always be the real vRefNum.
	
	WDPBRec myWDBRec;
	OSErr value_MyHGetVol;

	myWDBRec.ioCompletion = NULL;
	myWDBRec.ioNamePtr = NULL;
	value_MyHGetVol = PBHGetVol(&myWDBRec, kFSAsynch);
	*vRefNum = myWDBRec.ioWDVRefNum;					//the real vRefNum
	*dirID = myWDBRec.ioWDDirID;
	return value_MyHGetVol;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Str63 *GetTempFileName(Str63 *value_GetTempFileName)
{
	//Build a name for a temporary file by adding a suffix (contained in a
	//resource) to the application's name.
	
	Str255 tempStr;
	Str255 Temp_Str255_1;

	GetIndString(tempStr, rStrMisc, sTempFName);		//build file name
	PStr_Set((unsigned char *)value_GetTempFileName, PStr_Concat(Temp_Str255_1, 
																gAppName, 
																tempStr, 
																NULL));
	return value_GetTempFileName;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Terminate(void)
{
	//Clean up the application and exit.	Delete the temporary file.
	
	OSErr err;
	Str63 Temp_Str63_1;

	err = HDelete(gAppVRefNum, gAppParID, GetTempFileName((Str63 *)Temp_Str63_1));
	ExitToShell();									//exit if no cancellation
} //Terminate

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DrawItemFrame(DialogPtr dlgPtr, short itemNo)
{
	//Draws a frame around the dialog item.
	
	Rect iRect;
	PenState ps;
	Handle iHandle;
	short iKind;

	GetDialogItem(dlgPtr, itemNo, &iKind, &iHandle, &iRect);
	GetPenState(&ps);
	PenSize(1, 1);
	FrameRect(&iRect);
	SetPenState(&ps);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OSErr PathNameFromDirID(long DirID, short vRefnum, StringPtr fullPathName)
{
	//Given a DirID and real vRefnum, this routine will create and return the
	//full pathname that it corresponds to. It does this by calling
	//PBGetCatInfo for the given directory, and finding out its name and the
	//DirID of its parent. It the performs the same operation on the parent,
	//sticking its name onto the beginning of the first directory. This whole
	//process is carried out until we have processed the root directory
	//(identified with a DirID of fsRtDirID.
	
	//SPECIAL NOTE: This routine is once check for A/UX and would add the "/"
	//delimiter.	A/UX 2.0 supports the ":" just like HFS, and 2.0 is the
	//officially supported version.	 So, we no longer need to special case the
	//code.	 An important comment that was originally in this routine should be
	//considered again.
	
	//HOWEVER, BECAUSE OF THIS DEPENDENCY ON THE IDIOSYNCRASIES OF FILE
	//SYSTEMS, GENERATING FULL PATHNAMES FOR OTHER THAN DISPLAY PURPOSES IS
	//DISCOURAGED; IT'S CHANGED IN THE PAST WHEN A/UX WAS IMPLEMENTED, AND IT
	//MAY CHANGE AGAIN IN THE FUTURE IF SUPPORT FOR OTHER FILE SYSTEMS SUCH AS
	//PRODOS, MS-DOS, OR OS/2 ARE ADDED.
	
	//The full path name cannot exceed 255 chars.		If this happens, an error
	//bdNamErr is returned.	 So be careful out there!
	
	Str255 dirName;
	Str255 fullName;
	CInfoPBRec myCInfo;
	OSErr err;
	Str255 Temp_Str255_1;

	err = noErr;
	PStr_Set(fullName, "\p");
	PStr_Set(dirName, "\p");
	myCInfo.hFileInfo.ioNamePtr = (StringPtr)dirName;
	myCInfo.dirInfo.ioDrParID = DirID;
	
	do {
	myCInfo.hFileInfo.ioVRefNum = vRefnum;
	myCInfo.dirInfo.ioFDirIndex = -1;				//-1 means use ioDrDirID
	myCInfo.dirInfo.ioDrDirID = myCInfo.dirInfo.ioDrParID;
	err = PBGetCatInfo(&myCInfo, !kFSAsynch);
	if (err == noErr) {
		PStr_Set(dirName, PStr_Concat(Temp_Str255_1, dirName, "\p:", 
									NULL));
		if (PStr_Length(dirName) + PStr_Length(fullName) > 255) 
		err = bdNamErr;								//too big to eat!
		else 
		PStr_Set(fullName, PStr_Concat(Temp_Str255_1, dirName, fullName, 
										NULL));
	}
	} while (!(myCInfo.dirInfo.ioDrDirID == fsRtDirID || err != noErr));
	
	PStr_Set(fullPathName, fullName);
	return err;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OSErr PathNameFromWD(long wdRefNum, StringPtr pathName)
{
	//Given an HFS working directory, this routine returns the full pathname
	//that it corresponds to. It does this by calling PBGetWDInfo to get the
	//real VRefNum and DirID of the working directory. It then calls
	//PathNameFromDirID, and returns its result.
	
	//BUG NOTE:	 The special case portion of this routine that worked around an
	//U/UX 1.1 bug has been removed.	This application only runs with A/UX 2.0.
	//The following comment has been left here for historical purposes only.
	
	//PBGetWDInfo has a bug under A/UX 1.1.	 If vRefNum is a real vRefNum and
	//not a wdRefNum, then it returns garbage.	Since A/UX has only one volume
	//(in the Macintosh sense) and only one root directory, this can occur only
	//when a file has been selected in the root directory (/). So we look for
	//this and hardcode the DirID and vRefNum.
	
	//SPECIAL NOTE: Make sure you have read and understood Tech Note 238.
	
	long dirID;
	long procID;
	short vRefNum;
	OSErr err;

	err = GetWDInfo(wdRefNum, &vRefNum, &dirID, &procID);
	if (err == noErr) 
	err = PathNameFromDirID(dirID, vRefNum, pathName);
	return err;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ShowSelection(StringPtr name, StringPtr msg)
{
	//This routine shows an Alert of the selected item (name) and an optional
	//message (msg) that is used as the file's type.
	
	short theItem;

	ParamText(name, msg, "\p", "\p");
	theItem = CenteredAlert(rShowSelectionAlert);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ShowCancelled(void)
{
	//This simply calls up the alert showing the user selected the cancel
	//button of Standard File.
	
	Str255 tStr;
	short theItem;

	GetIndString(tStr, rStrMisc, sSelectedCancel);
	ParamText(tStr, "\p", "\p", "\p");
	theItem = CenteredAlert(rShowSelectionAlert);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OSErr ShowFoundFile(Str63 *fName, long parID, short vRefNum)
{
	//This alert shows the file name and the modification date to prove that the
	//file can be located by the given parameters.
	
	Str255 msg1;
	HParamBlockRec fHPBRec;
	short theItem;
	OSErr err;

	fHPBRec.ioParam.ioCompletion = NULL;
	fHPBRec.ioParam.ioNamePtr = (StringPtr)fName;
	fHPBRec.fileParam.ioFDirIndex = 0;
	fHPBRec.fileParam.ioDirID = parID;
	fHPBRec.ioParam.ioVRefNum = vRefNum;
	err = PBHGetFInfo(&fHPBRec, !kFSAsynch);
	if (err == noErr) {
	IUDateString(fHPBRec.fileParam.ioFlMdDat, longDate, msg1);
	ParamText(fName, msg1, "\p", "\p");
	theItem = CenteredAlert(rFInfoAlert);
	}
	return err;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//Dialog user item procedure to draw the multi-file list.	 The multi-file
//list item is requrest to be drawn by the Dialog Manager.	A handle to the
//list must be stored in the global variable "gListHandle" because the
//Dialog Manager isnt kind enough to let us pass a little information to
//our user-item procedures in the good computer science way, dagnabbit. The
//"theWindow" parameter is actually a pointer to our custom SFGetFile
//dialog. The "item" parameter specifies the item number to be drawn.
//Since only one item in our dialog uses UIPFileList, we dont really have
//to check to see if the item number is in fact the multi-file list, but we
//might as well so that people think were some cool programmers who take
//care to check everything.

//Check to be absolutely sure were talking about the file list item before
//we do anything to it. This alert shows the file name and the modification
//date to prove that the file can be located by the given parameters.

void UIPFileList(WindowPtr theWindow, short item)
{
	if (item == kFileListItem) {
	LUpdate(theWindow->visRgn, gListHandle);		//Draw visible part of list
	DrawItemFrame(theWindow, kFileListItem);		//...and frame it
	}
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void CreateFGetList(DialogPtr dlgPtr, RectPtr listRect, ListHandle *listH)
{
	//Build a three column list.	Set the column's width to cover the entire
	//list's visible rectangle.	 This is done by adjusting the cellSize.	This
	//keeps the other two columns hidden from the user.	 I also do not allow a
	//horizontal scrolling to prevent the user from viewing the hidden data.
	//It's a sick hack.	 I really don't like this idea, nor would I encourage
	//anyone else to do this.		Sorta, "Do as I say, not as I do."		I also
	//adjust the list rectangle for best scrolling.	 This is something that few
	//people do. They must be hard coding a size.		The adjusted rectangle is
	//returned so the caller will know the new size.
	
	Rect lBounds;										//Rectangle of list data
	Point cellSize;									//size of a cell
	FontInfo info;
	short newHeight;									//adjustement to list rect size

	InsetRect(listRect, kListFrameInset, kListFrameInset);
	listRect->right = listRect->right - kScrollbarAdjust;
	GetFontInfo(&info);
	cellSize.h = listRect->right - listRect->left;
	cellSize.v = info.ascent + info.descent + info.leading;
	newHeight = cellSize.v;
	while (newHeight + cellSize.v < listRect->bottom - listRect->top) 
	newHeight += cellSize.v;
	listRect->bottom = listRect->top + newHeight;
	SetRect(&lBounds, 0, 0, 3, 0);					//three columns in list
	*listH = LNew(listRect,							//Location of list in window
				&lBounds,							//Initial size of array
				cellSize,							//Cell size (0 = default)
				0,									//Resource ID of LDEF
				dlgPtr,								//Pointer to our Std File dialog
				true,								//needs to be drawn?
				false,								//grow box?
				false,								//horizontal scrolling?
				true);								//vertical scrolling?
	InsetRect(listRect, -kListFrameInset, -kListFrameInset);
	listRect->right = listRect->right + kScrollbarAdjust;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DeselectAllCells(ListHandle listH)
{
	Cell listCell;
	Boolean done;

	listCell = (Point)0;
	done = false;
	do 
	if (LGetSelect(true, &listCell, listH)) 
		LSetSelect(false, listCell, listH);
	else 
		done = true;
	while (!(done));
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void AddFItem(StringPtr name, long curDir, short vRefNum, ListHandle listH)
{
	//This takes the given information and adds it to the list.	 First thing to
	//do is check for the file to have already been added to the list.	If it
	//has, then we only have to show the user that the item is already there.
	//Before we eixt, we select just the current cell and the scroll it into view.
	
	Cell listCell;									//File list cell
	long itemDirID;
	short itemVRefNum;
	short dataLength;
	Boolean done;

	listCell = (Point)0;
	done = false;
	do 
	if (LSearch((Ptr)name[1], PStr_Length(name), NULL, &listCell, 
				listH)) {
		listCell.h = 1;
		dataLength = sizeof(itemDirID);
		LGetCell((Ptr)&itemDirID, &dataLength, listCell, listH);
		listCell.h = 2;
		dataLength = sizeof(itemVRefNum);
		LGetCell((Ptr)&itemVRefNum, &dataLength, listCell, listH);
		if (curDir == itemDirID && vRefNum == itemVRefNum) {
		done = true;								//found the existing item
		}
	} else {
		listCell.h = 0;
		listCell.v = LAddRow(1, (*listH)->dataBounds.bottom, listH);
		LSetCell((Ptr)&gReply.fName[1], PStr_Length(name), listCell, 
				listH);
		listCell.h = 1;
		LSetCell((Ptr)&curDir, sizeof(curDir), listCell, listH);
		listCell.h = 2;
		LSetCell((Ptr)&vRefNum, sizeof(vRefNum), listCell, listH);
		LSetSelect(true, listCell, listH);
		done = true;
	}
	while (!(done));
	DeselectAllCells(listH);
	listCell.h = 0;									//first column, current row
	LSetSelect(true, listCell, listH);
	LAutoScroll(listH);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void GetFItems(ListHandle listH)
{
	//We need to copy all of the files from the global list and store these into
	//a global array of file descriptions.
	
	Point theCell;
	short numOfFiles;
	short dataLength;
	short index;

	numOfFiles = (*listH)->dataBounds.bottom;
	gMultiFiles = (FileSpecHdl)NewHandle(sizeof(FileDesc) * numOfFiles);
	MoveHHi((Handle)gMultiFiles);						//helps to reduce fragmentation
	HLock((Handle)gMultiFiles);
	for (index = 1; index <= numOfFiles; ++index) {
	dataLength = sizeof((*(*gMultiFiles))[index - 1].vRefNum);
	SetPt(&theCell, 2, index - 1);
	LGetCell((Ptr)&(*(*gMultiFiles))[index - 1].vRefNum, &dataLength, 
				theCell, listH);
	dataLength = sizeof((*(*gMultiFiles))[index - 1].parID);
	SetPt(&theCell, 1, index - 1);
	LGetCell((Ptr)&(*(*gMultiFiles))[index - 1].parID, &dataLength, 
				theCell, listH);
	dataLength = sizeof((*(*gMultiFiles))[index - 1].name);
	SetPt(&theCell, 0, index - 1);
	LGetCell((Ptr)&(*(*gMultiFiles))[index - 1].name[1], &dataLength, 
				theCell, listH);
	(*(*gMultiFiles))[index - 1].name[0] = (char)dataLength;
	}
	HUnlock((Handle)gMultiFiles);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void RemoveFItems(ListHandle listH)
{
	//Delete all selected files from the multi-file list.		Dont want all that
	//flickerin, so turn off drawing.	Keep deleting selected rows until no
	//selected rows are left.		Finally, turn on drawing and erase the list.
	//Itll get drawn on the next update event.	 This takes the given
	//information and adds it to the list.
	
	Rect listRect;									//Rectangle for list
	Cell listCell;									//File list cell

	LSetDrawingMode(false, listH);
	listCell = (Point)0;
	while (LGetSelect(true, &listCell, listH)) {
	LDelRow(1, listCell.v, listH);
	listCell = (Point)0;
	}
	listCell = (Point)0;
	LSetSelect(true, listCell, listH);
	LAutoScroll(listH);
	listRect = (*listH)->rView;
	LSetDrawingMode(true, listH);
	InvalRect(&listRect);
	EraseRect(&listRect);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void CreateFPutList(DialogPtr dlgPtr, RectPtr listRect, ListHandle *listH, StringPtr cellStr)
{
	//This creates a list used to hold a number of file types.	The selected
	//item is used to determine which format the file should be saved as.		The
	//list rectangle is adjusted to the best scrolling size, and the adjusted
	//size is returned to the caller.		It's also important to select the first
	//item in the list as a default file format and that only one item can be
	//selected.
	
	Str255 entry;
	Rect lBounds;
	Point newSize;
	Point theCell;
	short newRow;
	short dataLen;

	InsetRect(listRect, kListFrameInset, kListFrameInset);
	listRect->right = listRect->right - kScrollbarAdjust;
	SetRect(&lBounds, 0, 0, 1, 0);					//one dimentional list
	*listH = LNew(listRect,							//position in window
				&lBounds,							//initial size of array
				(Point)0,							//cell size (0 = default)
				0,									//resource ID of LDEF
				dlgPtr,								//window pointer
				true,								//drawit
				false,								//has grow
				false,								//scrollHoriz
				true);								//scrollVert
	newSize = (*(*listH))->cellSize;					//get the size of one cell
	while (newSize.v + (*(*listH))->cellSize.v < listRect->bottom - listRect->top) 
	newSize.v = newSize.v + (*(*listH))->cellSize.v;
	LSize(newSize.h, newSize.v, *listH);				//adjust for best scrolling
	(*(*listH))->selFlags = lOnlyOne;					//single selections only
	
	theCell = (Point)0;
	do {
	GetIndString(entry, rStrList, theCell.v + 1);
	if (PStr_Cmp(entry, "\p") != 0) {
		newRow = LAddRow(1, (*(*listH))->dataBounds.bottom, *listH);
		LSetCell((Ptr)&entry[1], PStr_Length(entry), theCell, *listH);
	}
	theCell.v = theCell.v + 1;
	} while (PStr_Cmp(entry, "\p") != 0);
	theCell = (Point)0;
	LSetSelect(true, theCell, *listH);
	dataLen = sizeof(cellStr);
	LGetCell((Ptr)cellStr[1], &dataLen, theCell, *listH);
	cellStr[0] = (char)dataLen;
	
	listRect->bottom = listRect->top + newSize.v;
	InsetRect(listRect, -kListFrameInset, -kListFrameInset);
	listRect->right = listRect->right + kScrollbarAdjust;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DisposeFPutList(ListHandle listH, StringPtr cellStr)
{
	//Just get rid of the list, but before we do that get the currently selected
	//cell's data which is the user's choice of file format.
	
	Point theCell;
	short dataLen;

	SetPt(&theCell, 0, 0);
	if (LGetSelect(true, &theCell, listH)) {
	dataLen = sizeof(cellStr);
	LGetCell((Ptr)cellStr[1], &dataLen, theCell, listH);
	cellStr[0] = (char)dataLen;						//length byte
	}
	LDispose(listH);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//Simple SFGetFile example
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//Simplest form of SFGetFile.	 This routine puts up a GetFile dialog with a
//request to show all files.	When this is done, a check is made to see
//which of the Save or Cancel buttons were pressed.		If the Save button was
//pressed, ShowSelection is called to display which file was selected.	If
//the Cancel button was pressed, a dialog is shown saying so.

//A full path name is shown of the selected file.	 This is done as an
//example.	Generally full path names should be avoided.	If you wanted to
//remember the location of a file, such as the next time the application is
//launch, you should remember the volume name, directory ID, and file name.
//As an example of how to do this, refer to the routine DoRememberFile.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ShowPath(void)
{
	Str255 pathStr;
	SFTypeList typeList;
	SFReply reply;
	OSErr err;
	Str255 Temp_Str255_1;

	SFGetFile((Point)kSFTopLeft,						//location
			"\p",									//prompt String
			NULL,									//file filter
			kShowAllFiles,							//numtypes
			typeList,								//array to types to show
			NULL,									//dialog hook
			&reply);								//record for returned values
	if (reply.good) {
	err = PathNameFromWD(reply.vRefNum, pathStr);
	if (err == noErr) {
		if (PStr_Length(pathStr) + PStr_Length(reply.fName) > 255) 
		AlertUser(bdNamErr, sFileSystem);
		else 
		ShowSelection(PStr_Concat(Temp_Str255_1, pathStr, reply.fName, 
									NULL), "\p");
	} else 
		AlertUser(err, sFileSystem);
	} else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//SFPGetFile with Dialog Hook and Simple File Filter
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//Depending on the value of the global variable "gNormal", this
//demonstrations shows either TEXT files or APPL files.		The value of
//"gNormal" is determined by the states of two radio buttons that are added
//to the dialog box.	Our dialog hook routine is used to initialize the radio
//buttons and handle hits on them.	When there is a hit on a radio button,
//"gNormal" is set to either TRUE or FALSE, and a special command is sent
//back to Standard File telling it to regenerate the list of file names it
//is displaying.	This sample consists of 3 parts:

//DoNormalPGet
//------------
//This routine initializes a variable for our file filter, and then calls
//SFPGetFile with pointers to two other routines and a resource number for
//a special dialog box with extra items in it.

//SimpleFFilter
//----------------
//Specified in our SFPGetFile call to be called to specify whether or not a
//file should be displayed.		If the global variable 'gNormal' is TRUE, then
//all text are displayed.	 Otherwise, Applications will be displayed.

//MySFGetDlgHook
//----------------
//A routine that is called to handle hits on the non-standard items in our
//dialog box.	 The dialog hook is also used to perform some special
//initialization on the items in the dialog box.	Standard file does this by
//calling this routine with a bogus 'item' number of -1.	When we get this
//number, we initialize our radio buttons, and change the text in the Open
//button.

//The radio buttons are used to determine what files will appear in the
//files list.	 It does this by appropriately setting the 'gNormal' variable
//for SimpleFFilter, and then telling Standard File that the file list
//needs to be regenerated by passing back 101 as the function result.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

short MySFGetDlgHook(short MySFitem, DialogPtr dlgPtr)
{
	short value_MySFGetDlgHook = MySFitem;

	switch (MySFitem) { 
	
	case kFirstTime:
		SetRadioButton(dlgPtr, kAppButton, kCntlOff);
		SetRadioButton(dlgPtr, kTextButton, kCntlOn);
		break;
	
	case kTextButton:
		if (!gNormal) {
		SetRadioButton(dlgPtr, kAppButton, kCntlOff);
		SetRadioButton(dlgPtr, kTextButton, kCntlOn);
		gNormal = true;
		value_MySFGetDlgHook = kReDrawList;
		}
		break;
	
	case kAppButton:
		if (gNormal) {
		SetRadioButton(dlgPtr, kTextButton, kCntlOff);
		SetRadioButton(dlgPtr, kAppButton, kCntlOn);
		gNormal = false;
		value_MySFGetDlgHook = kReDrawList;
		}
		break;
	}
	
	return value_MySFGetDlgHook;
}

Boolean SimpleFFilter(ParmBlkPtr p)
{
	Boolean value_SimpleFFilter = !kShowIt;

	if (gNormal) {
	if (p->fileParam.ioFlFndrInfo.fdType == 'TEXT') 
		value_SimpleFFilter = kShowIt;
	} else if (p->fileParam.ioFlFndrInfo.fdType == 'APPL') 
	value_SimpleFFilter = kShowIt;
	return value_SimpleFFilter;
}

void DoNormalPGet(void)
{
	SFTypeList typeList;
	SFReply reply;
	OSErr err;
	FileFilterUPP ffUPP;
	DlgHookUPP dhUPP;

	gNormal = true;
	
	ffUPP = NewFileFilterProc((FileFilterProcPtr)SimpleFFilter);
	dhUPP = NewDlgHookProc((DlgHookProcPtr)MySFGetDlgHook);
	SFPGetFile((Point)kSFTopLeft,						//location
				"\p",									//vestigial String
				ffUPP,									//file filter
				kShowAllFiles,							//numtypes
				typeList,								//array to types to show
				dhUPP,									//dialog hook
				&reply,								//record for returned values
				rSFPGetFileDLOG,						//ID of Custom Dialog
				NULL);									//ModalDialog filterProc
	DisposeRoutineDescriptor(dhUPP);
	DisposeRoutineDescriptor(ffUPP);
	if (reply.good) 
	ShowSelection(reply.fName, "\p");
	else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//SFGetFile with a complex File Filter
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//Sample of a less trivial file filter.		This routine is responsible for
//displaying only files that have 'ALRT' resources.		It shows how to
//correctly look inside a file for qualifying information, and demonstrates
//how to get around a particular quirk when attempting multiple access to
//the resource fork.

//DoFFilter
//---------
//Calls SFGetFile and specifies the File Filter.

//ComplexFFilter
//--------------
//This routine used to do skanky things like looking at a low memory global
//to determine of a new resource file map was opened or not.	Even more
//nastiness was the it created a working directory.		Well, this was all too
//sick and a completely different approach is now taken.	Opening a
//resource file as read-only will always create a new map.	This should
//generally be avoided, but here I'm opening it and then immediately
//closing it without allowing anything else to access the file.

//Opening resource forks can be tricky if it's already open.	It would be
//very bad to use a resource fork that is already open by another
//application.	The problem being that the Resource Manager doesn't deal
//with multiple users.	Read Tech notes #116 and #185.	There's also
//another problem.	If a resource fork is opened by two applications and
//one closes the file then the entire resource fork may closed out from
//underneath the other application.		I do, however, want to show the user
//files that contain ALRT resources even if the file is currently open.		I
//use HOpenResFile with read only permission, which will give me a unique
//resource reference.	 I look for an ALRT resource and then immediately
//close the file.	 Do not use a resource file opened with read only
//permission because another application could come along and start writing
//to it.	Another reason to use HOpenResFile	is to avoid a necessary
//working directory.	I can use the DirID kept updated by Standard File.
//Performing this search on each file is time consuming so I show a
//spinning cursor to show the user I'm working.		Opening a resource fork
//may load resources mark preload.	To avoid this, I call SetResLoad to
//FALSE.	Read Tech Note #203 for other reasons not to play with resources.


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Boolean ComplexFFilter(ParmBlkPtr p)
{
	//SPECIAL NOTE: The ParamBlockRec is a relocatable object in the heap.	The
	//file name is one of the fields in this handle.	Because of this we need
	//to copy the fields that we use to a local variable to avoid deferenced
	//handle problems.
	
	Str255 fileName;
	long oldTicks;
	short resRef;
	short curRes;
	short vRefNum;
	Boolean value_ComplexFFilter;

	oldTicks = TickCount();
	value_ComplexFFilter = !kShowIt;					//dont show anything until I say so
	PStr_Set(fileName, p->ioParam.ioNamePtr);			//must save name
	vRefNum = p->ioParam.ioVRefNum;					//must save volume (or WD) reference
	curRes = CurResFile();
	SetResLoad(false);
	resRef = HOpenResFile(vRefNum, GetSFCurDir(), fileName, fsRdPerm);
	if (resRef != -1) {
	UseResFile(resRef);
	if (Count1Resources('ALRT') > 0) 
		value_ComplexFFilter = kShowIt;				//hey, we found an Alert in here
	CloseResFile(resRef);
	}													//restore everything
	SetResLoad(true);
	UseResFile(curRes);
	RotateCursor(TickCount() - oldTicks);
	return value_ComplexFFilter;
}

void DoFFilter(void)
{
	SFTypeList typeList;
	SFReply reply;

	SpinCursor(0);									//get the spinning cursor ready
	SFGetFile((Point)kSFTopLeft,						//location
			"\p",									//vestigial String
			(FileFilterUPP)ComplexFFilter,			//file filter
			kShowAllFiles,							//numtypes
			typeList,								//array to types to show
			NULL,									//dialog hook
			&reply);								//record for returned values
	if (reply.good) 
	ShowSelection(reply.fName, "\p");
	else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//DoGetDirectory with a Dialog Hook and File Filter
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This demonstrates how to modify SFGetFile to allow you to select a
//directory.	This mimics the "GetFileName -d" function of MPW.		As a matter
//of fact, the DLOG and DITL used in this sample were taken directly out of
//MPW.	There are 2 major additions in the dialog used in this sample:

//1. a Simple button that lets one select the directory that is currently
//	highlighted in the list of directories,
//2. Simple button at the top of the dialog that lets the user select the
//	directory that we are currently *IN*.

//DoGetDirectory
//--------------
//Sets up the pointers to the Dialog Hook and File Filter.	Once the user
//has selected a folder, we look at the global gMyCurDir which was set
//within the Dialog Hook.	 We take this DirID and call PBGetCatInfo to find
//the name of this folder.	This is so we can display the name to the user.

//MyGetDirDlgHook
//----------------
//No new techniques are really used in this sample.		Hits on the two simple
//buttons are handled by a dialog hook called GetDirDlgHook.	Depending on
//which button is hit, we set the global variable gMyCurDir to either
//gReply.fType (for the currently highlighted directory) or what Standard
//File has set as the current directory.	We then simulate a hit on the Open
//button so that Standard File will return to our application and the
//FSReply.good will be TRUE.

//While an item is selected, the gReply.fType will be set to a value.
//gReply.fType is set to 0 when nothing is selected.	We test for this
//during idle time (item = 100) and activate the Directory button
//appropriately.	We also use the "Item = -1" feature to set up a prompt
//String.

//FolderFFilter
//-------------
//Normally, folders are ALWAYS shown, and aren't even passed to this file
//filter for judgement.		Under such circumstances, it is only necessary to
//blindly return TRUE (allow no files whatsoever).	However, Standard File
//is not documented in such a manner, and this feature may not be TRUE in
//the future.	 Therefore, we DO check to see if the entry passed to us
//describes a file or a directory.	This can be done by checking a bit
//in the ioFlAttrib, which is set for folders.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Boolean FolderFFilter(ParmBlkPtr p)
{
	return !BTst(p->fileParam.ioFlAttrib, kFolderBit);
}

short GetDirDlgHook(short item, DialogPtr dlgPtr)
{
	Str255 messageTitle;
	Rect iRect;
	Handle iHandle;
	short iKind;
	short value_GetDirDlgHook = item;

	switch (item) { 
	case kFirstTime:								//Read prompt String from resource
		GetIndString(messageTitle, rStrMisc, sSelectFolder);
		GetDialogItem(dlgPtr, kGetDirMessage, &iKind, &iHandle, &iRect);
		SetDialogItemText(iHandle, messageTitle);
		break;
	
	case kGetDirButton:
		if ((long)gReply.fType != 0) {
		gMyCurDir = (long)gReply.fType;
		value_GetDirDlgHook = getOpen;				//simulate Open button
		}
		break;
	
	case kGetDirNowButton:
		gMyCurDir = GetSFCurDir();
		value_GetDirDlgHook = getOpen;				//simulate Open button
		break;
	
	case kNullModalEvt:
		if ((long)gReply.fType == kNoSelection) {
		GetDialogItem(dlgPtr, kGetDirButton, &iKind, &iHandle, &iRect);
		HiliteControl((ControlHandle)iHandle, kCntlDeactivate);
		} else {
		GetDialogItem(dlgPtr, kGetDirButton, &iKind, &iHandle, &iRect);
		HiliteControl((ControlHandle)iHandle, kCntlActivate);
		}
		break;
	}
	
	return value_GetDirDlgHook;
}

void DoGetDirectory(void)
{
	Str31 folderName;									//directories can be 31 chars
	SFTypeList typeList;
	CInfoPBRec myCInfo;
	OSErr err;
	FileFilterUPP ffUPP;
	DlgHookUPP dhUPP;

	ffUPP = NewFileFilterProc((FileFilterProcPtr)FolderFFilter);
	dhUPP = NewDlgHookProc((DlgHookProcPtr)GetDirDlgHook);
	SFPGetFile((Point)kSFTopLeft,						//location
				"\p",									//vestigial String
				ffUPP,									//file filter
				kShowAllFiles,							//numtypes
				typeList,								//array to types to show
				dhUPP,									//Dialog hook
				&gReply,								//record for returned values
				rGetDirectoryDLOG,						//Resource ID dialog
				NULL);									//ModalDialog filter proc
	DisposeRoutineDescriptor(dhUPP);
	DisposeRoutineDescriptor(ffUPP);
	
	if (gReply.good) {
	PStr_Set(folderName, "\p");						//initialize string
	myCInfo.hFileInfo.ioCompletion = NULL;			//no completion
	myCInfo.hFileInfo.ioNamePtr = (StringPtr)folderName;	//VAR to string
	myCInfo.dirInfo.ioFDirIndex = -1;				//-1 means use ioDrDirID
	myCInfo.hFileInfo.ioVRefNum = GetSFVRefNum();	 //this is the volume
	myCInfo.dirInfo.ioDrDirID = gMyCurDir;			//this is the directory
	err = PBGetCatInfo(&myCInfo, !kFSAsynch);
	if (err == noErr) 
		ShowSelection(folderName, "\p");
	else 
		AlertUser(err, sFileSystem);
	} else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//DoMultiFile
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This demonstrates how to get multiple files with one call to Standard
//File.		It collects a set of files while in the SFGetFile dialog by using
//of a secondary list.	This is similar to the method used by many of the
//MPW tools.	The original version of this example was a failed experiment.
//It attempted remember all of the files passed to the file filter, and
//using those names to create a new, application governed file list that
//will replace the one displayed by Standard File.	However, this
//experiment seems doomed to failure, for the following reasons:

//* Icons are not displayed next to the names of files.		This would require
// a custom LDEF.	 The resulting presentation without the small icons was
// more disconcerting that I thought.
//* The current directory button (the one displayed above the list of file
// names) disappears.	 This seems due to the fact that the pop-up menu-like
// item is not actually a dialog item, and is somehow attached to the list
// of file names handled by Standard File.	Since I move that list off of the
// window so that I can display my own, the current directory button seems
// to go with it.
//* Updating the list is difficult.		The normal sequence of events would
// desirably be: 1) have the file filter create a linked list of names to
// appear in the dialog, 2) insert those names into the List Manager list,
// and 3) display the list.		However, by following this outline, the list
// does not get updated, as the update region for that area is empty.	 So, at
// some time, we need to invalidate the area that holds our list.	 But there
// is no convenient place in which to do that.	We can't call InvalRect
// within our file filter routine, as we don't have the DialogPtr at the
// time (needed to get the bounding rectangle of the list item).	Neither can
// we call InvalRect within step 2, as we are between
// BeginUpdate/EndUpdates, and calling InvalRect would be useless.
//* We can't display directories.	 They are not offered to the judgement of
// our file filter, and hence cannot be added to our list.
//* Key presses would have do be handled manually by our dialog hook.

//Considering all of these problems, another possible solution would be to
//do away with Standard File altogether and do your own non-Standard File.
//You could show a dialog (maybe even a modeless dialog) and contained a
//list of file in the selected folder.	This really isn't all too
//difficult.	The only disadvantage is that it may not look familiar to the
//user, and the MultiFinder hack to force the application's Open item
//wouldn't work.	This process is documented in Tech Note 205.

//DoMultiFile
//-----------
//Points to the Dialog Filter and Dialog Hook.	If the user selected the
//Open button then there is a handle containing an array of files.	We
//index through this array and show some information for each file proving
//that we located the file.		Once all the files have been found, then
//dispose of the handle.

//ListGetDlgHook
//--------------
//This routine is at the dialog item level, rather than the event level.
//Normally, double-clicking on a file name in the Standard File list
//simulates a click in the Open button.		We want it to add an item to our
//multi-file list instead, so we've turned the Open button into an Add
//button and sabotaged it's functionality so that it doesn't dismiss the
//Standard File dialog.		Then, we added our own Open button which tricks
//the Standard File code into thinking the normal Open button was clicked,
//and our dialog is dismissed.

//If there was an event (item <> 100) then handle the activating of our
//Open button and the Remove buttons.	 Activate the Open button if there
//are any files in our file list.	 Activate the Remove button if there are
//any selected files in our file list.	If the Open or Cancel buttons were
//selected, then dispose of the list since the user has dismissed the
//dialog.

//FListDlgFilter
//--------------
//This routine's raison d'tre is to handle clicks in our file list (that
//is, OUR file list at the bottom of the dialog; not SF's file list at the
//top of the dialog).	 If FListDlgFilter weren't around, clicks in our file
//list would be ignored.	First we check to see if the mouse was clicked
//somewhere in the rectangle of our file list.	If it wasn't, we return
//FALSE, and SF handles the event on it's own.	If it was, call LClick and
//let the List Manager handle all the hilighting and scrolling.

//If there was a mouse click in our file list item, then itemHit returns
//the item number of our file list.		If FListDlgFilter found out that the
//mouse click wasn't in our file list, then itemHit is unchanged and we
//return FALSE which tells the Dialog Manager to figures out which item has
//been affected by the latest event.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Boolean FListDlgFilter(DialogPtr dlgPtr, EventRecord *evt, SInt16 *itemHit)
{
	Rect iRect;
	Point clickPos;
	Handle iHandle;
	short iKind;
	Boolean ignore;
	Boolean value_FListDlgFilter = false;

	if (evt->what == mouseDown) {						//Only care about mouseDown in our list
	clickPos = evt->where;
	GlobalToLocal(&clickPos);
	GetDialogItem(dlgPtr, kFileListItem, &iKind, &iHandle, &iRect);
	if (PtInRect(clickPos, &iRect)) {
		ignore = LClick(clickPos, evt->modifiers, gListHandle);
		*itemHit = kFileListItem;
		value_FListDlgFilter = true;
	}
	}
	return value_FListDlgFilter;
}

short ListGetDlgHook(short item, DialogPtr dlgPtr)
{
	Rect iRect;
	Handle iHandle;
	short iKind;
	Cell listCell;									//File list cell
	short value_ListGetDlgHook = item;

	switch (item) { 
	
	case kFirstTime:
		GetDialogItem(dlgPtr, kMultiOpenItem, &iKind, &iHandle, &iRect);
		HiliteControl((ControlHandle)iHandle, kCntlDeactivate);
		GetDialogItem(dlgPtr, kRemoveItem, &iKind, &iHandle, &iRect);
		HiliteControl((ControlHandle)iHandle, kCntlDeactivate);
		GetDialogItem(dlgPtr, kFileListItem, &iKind, &iHandle, &iRect);
		CreateFGetList(dlgPtr, &iRect, &gListHandle);
		SetDialogItem(dlgPtr, kFileListItem, iKind, (Handle)UIPFileList, 
					&iRect);
		break;										//kFirstTime
	
	case getOpen:									//old Open button means add a file
		AddFItem(gReply.fName, GetSFCurDir(), GetSFVRefNum(), gListHandle);
		value_ListGetDlgHook = kNoItem;				//replace item with no item
		break;
	
	case kMultiOpenItem:							//new Open button means get all files
		GetFItems(gListHandle);
		value_ListGetDlgHook = getOpen;				//replace item with real Open item
		break;
	
	case kRemoveItem:
		RemoveFItems(gListHandle);					//remove the selected item from list
		break;
	}
	
	//CASE
	
	if (item != kNullModalEvt) {
	GetDialogItem(dlgPtr, kMultiOpenItem, &iKind, &iHandle, &iRect);
	ListPtr with_3 = *gListHandle;
	if (with_3->dataBounds.top != with_3->dataBounds.bottom) 
		HiliteControl((ControlHandle)iHandle, kCntlActivate);
	else 
		HiliteControl((ControlHandle)iHandle, kCntlDeactivate);
	listCell = (Point)0;
	GetDialogItem(dlgPtr, kRemoveItem, &iKind, &iHandle, &iRect);
	if (LGetSelect(true, &listCell, gListHandle)) 
		HiliteControl((ControlHandle)iHandle, kCntlActivate);
	else 
		HiliteControl((ControlHandle)iHandle, kCntlDeactivate);
	}
	
	if (item == kMultiOpenItem || item == getCancel) 
	LDispose(gListHandle);
	return value_ListGetDlgHook;
}

void DoMultiFile(void)
{
	Str255 msg1;
	SFTypeList typeList;
	short index;
	short numOfFiles;
	HParamBlockRec fHPBRec;
	OSErr err;
	short theItem;
	DlgHookUPP dhUPP;
	ModalFilterUPP mfUPP;

	dhUPP = NewDlgHookProc((DlgHookProcPtr)ListGetDlgHook);
	mfUPP = NewModalFilterProc((ModalFilterProcPtr)FListDlgFilter);
	SFPGetFile((Point)kSFTopLeft,						//Location of SFDialog
				"\p",									//Heres that String thing.
				NULL,									//File filter
				kShowAllFiles,							//Number of file types to display
				typeList,								//Array of file types to show
				dhUPP,									//Std File dialog hook
				&gReply,								//Record for returned SF info
				rListGetDLOG,							//Resource ID of dialog
				mfUPP);								//ModalDialog filter proc
	DisposeRoutineDescriptor(dhUPP);
	DisposeRoutineDescriptor(mfUPP);
	
	if (gReply.good) {
	numOfFiles = GetHandleSize((Handle)gMultiFiles) / sizeof(FileDesc);
	
	HLock((Handle)gMultiFiles);
	for (index = 1; index <= numOfFiles; ++index) {
													//prepare a hParamBlock
		fHPBRec.ioParam.ioCompletion = NULL;
		fHPBRec.ioParam.ioNamePtr = (StringPtr)(*(*gMultiFiles))[index - 1].name;
		fHPBRec.ioParam.ioVRefNum = (*(*gMultiFiles))[index - 1].vRefNum;
		fHPBRec.fileParam.ioFDirIndex = 0;
		fHPBRec.fileParam.ioDirID = (*(*gMultiFiles))[index - 1].parID;
		err = PBHGetFInfo(&fHPBRec, !kFSAsynch);		//fPBRec on stack, synch only
		
		if (err == noErr) {
		IUDateString(fHPBRec.fileParam.ioFlMdDat, longDate, msg1);
		ParamText((*(*gMultiFiles))[index - 1].name, msg1, "\p", "\p");
		theItem = CenteredAlert(rFInfoAlert);
		} else 
		AlertUser(err, sFileSystem);
	}
	DisposeHandle((Handle)gMultiFiles);
	} else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//Typical SFPutFile example
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//Simplest form of SFPutFile. This routine puts up a PutFile dialog with a
//prompt and suggested file name.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DoNormalPut(void)
{
	Str255 msg;										//prompt String
	SFReply reply;

	GetIndString(msg, rStrMisc, sSaveAsMsg);
	SFPutFile((Point)kSFTopLeft,						//location
			msg,									//prompt String
			"\pDoug",								//original name
			NULL,									//dialog hook
			&reply);								//record for returned values
	if (reply.good) 
	ShowSelection(reply.fName, "\p");
	else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//SFPPutFile with a Dialog Hook
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This demonstration is a custom PutFile routine which adds a few items to
//the dialog box.	 There are two radio buttons that determine what format
//the user wants to save the information in.	And there is a Static Text
//item that displays the amount of free space left on the disk.		This text
//item is updated whenever we detect that the user has switched to another
//volume.

//DoNormalPPut
//------------
//This routine calls SFPPutFile.	Note the extra 'P' in the name.	 This
//Toolbox call allows us to specify a customized dialog box, and dialog
//hook that handles hits on special items.	There is a global string used
//to show what the file format will be to the user.

//MySFPutDlgHook
//--------------
//A routine that is called to handle the non-standard items in our dialog
//box.	These items are:

//* The control of two radio buttons that affect the setting of a global
// variable used to determine what format the file will be saved in
//* UserItem that displays the amount of free space on the current volume.

//The dialog hook is also used to perform some special initialization on
//the items in the dialog box.	We initialize our radio buttons, change the
//text in the save button, and prepare the user item by initializing the
//routine that will draw it and getting the text that will appear in it.

//Finally, this routine will force the user item to be drawn once at
//initialization time, and when ever the user clicks the Drive button.	By
//invalidating the user item rectangle, it causes the drawing procedure to
//update according to the current volume which is set by Standard File.

//DrawFreeSpace
//-------------
//This is a user item's drawing procedure.	It is used to display the
//amount of free space left on the disk.	All we need to do is update the
//text in our userItem.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DrawFreeSpace(WindowPtr theWindow, short itemNo)
{
	TwoIntsMakesALong convert;
	Str255 myStr1;
	Str255 myStr2;
	Handle iHandle;
	short iKind;
	Rect iRect;
	OSErr err;
	long space;
	Str255 Temp_Str255_1;

	space = GetKFreeSpace(GetSFVRefNum());
	if (space < 0) 
	PStr_Set(myStr1, "\p??");						//we couldn't get the free space size!
	else 
	NumToString(space, myStr1);
	GetIndString(myStr2, rStrMisc, sKFree);			//Text to be added in our user item.
	PStr_Set(myStr1, PStr_Concat(Temp_Str255_1, myStr1, myStr2, NULL));
	GetDialogItem((DialogPtr)theWindow, kSFPPutFreeSpace, &iKind, &iHandle, 
				&iRect);
	TETextBox((Ptr)&myStr1[1], PStr_Length(myStr1), &iRect, teJustCenter);
}

short MySFPutDlgHook(short item, DialogPtr dlgPtr)
{
	Str255 saveStr;									//replacement name for Save button
	Rect iRect;
	Handle iHandle;
	short iKind;
	short value_MySFPutDlgHook = item;

	switch (item) { 
	
	case kFirstTime:
		SetRadioButton(dlgPtr, kSFPPutTextBtn, kCntlOff);
		GetDialogItem(dlgPtr, kSFPPutNormalBtn, &iKind, &iHandle, &iRect);
		SetControlValue((ControlHandle)iHandle, kCntlOn);
		GetControlTitle((ControlHandle)iHandle, gPutFormat);
		GetDialogItem(dlgPtr, kSFPPutFreeSpace, &iKind, &iHandle, &iRect);
		SetDialogItem(dlgPtr, kSFPPutFreeSpace, iKind, (Handle)DrawFreeSpace, 
					&iRect);
		InvalRect(&iRect);							//force item to update
		break;
	
	case getDrive:
		GetDialogItem(dlgPtr, kSFPPutFreeSpace, &iKind, &iHandle, &iRect);
		InvalRect(&iRect);
		break;
	
	case kSFPPutNormalBtn:
		if (!gNormal) {
		SetRadioButton(dlgPtr, kSFPPutTextBtn, kCntlOff);
		GetDialogItem(dlgPtr, kSFPPutNormalBtn, &iKind, &iHandle, 
						&iRect);
		SetControlValue((ControlHandle)iHandle, kCntlOn);
		GetControlTitle((ControlHandle)iHandle, gPutFormat);
		gNormal = true;								//change our flag accordingly
		}
		break;
	
	case kSFPPutTextBtn:
		if (gNormal) {
		SetRadioButton(dlgPtr, kSFPPutNormalBtn, kCntlOff);
		GetDialogItem(dlgPtr, kSFPPutTextBtn, &iKind, &iHandle, &iRect);
		SetControlValue((ControlHandle)iHandle, kCntlOn);
		GetControlTitle((ControlHandle)iHandle, gPutFormat);
		gNormal = false;							//change our flag accordingly
		}
		break;
	}
	
	//CASE
	return value_MySFPutDlgHook;
}

void DoNormalPPut(void)
{
	Str255 msg;										//prompt String
	SFReply reply;
	DlgHookUPP dhUPP;

	gNormal = true;
	GetIndString(msg, rStrMisc, sSaveAsMsg);
	
	dhUPP = NewDlgHookProc((DlgHookProcPtr)MySFPutDlgHook);
	SFPPutFile((Point)kSFTopLeft,						//location
				msg,									//prompt String
				"\pDoug",								//original name
				dhUPP,									//dialog hook
				&reply,								//record for returned values
				rSFPPutFileDLOG,						//ID of custom Dialog
				NULL);									//ModalDialog filterProc
	DisposeRoutineDescriptor(dhUPP);
	
	if (reply.good) 
	ShowSelection(reply.fName, gPutFormat);
	else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//DoForceDirectory
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This is a quick sample that shows how to set the initial directory that
//Standard File comes up with.	Basically, this is done by storing
//appropriate values into SFSaveDisk and CurDirStore.	 In this example, I
//force the directory to be the System Folder of the boot volume as
//specified by SysEnvirons.		Since I like to be user friendly, I save and
//restore the current folder.	 So the next time Standard File is called it
//will be exactly where the user left it.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DoForceDirectory(void)
{
	Str255 msg;										//prompt String
	SFTypeList typeList;
	long oldDirStore;
	long dirID;
	long procID;
	short vRefNum;
	short oldVRefNum;
	SFReply reply;
	OSErr err;

	oldDirStore = GetSFCurDir();						//save old folder and volume
	oldVRefNum = GetSFVRefNum();
	err = GetWDInfo(gMac.sysVRefNum, &vRefNum, &dirID, &procID);
	if (err == noErr) {
	SetSFCurDir(dirID);
	SetSFVRefNum(vRefNum);
	GetIndString(msg, rStrMisc, sSaveAsMsg);
	SFPutFile((Point)kSFTopLeft,					//location
				msg,									//prompt String
				"\pDoug",								//original name
				NULL,									//dialog hook
				&reply);								//record for returned values
	if (reply.good) 
		ShowSelection(reply.fName, "\p");
	else 
		ShowCancelled();
	SetSFCurDir(oldDirStore);						//restore old folder and volume
	SetSFVRefNum(oldVRefNum);
	} else 
	AlertUser(err, sStandardErr);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//DoPutListsFile
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This sample demonstrates putting a List Manager list into an SFPPutFile
//dialog, describes a problem that can occur by doing so, and shows one
//solution to the problem.	Putting another list into a Standard File
//dialog can be useful for a number of reasons.		While *I* can't think of a
//reason I would want to put a second list into a PutFile dialog, it is
//possible that you may wish to.	The best example that I could think of is
//a list of file types for the user to choose.	However, there is a subtle
//problem that arises if you do: in the following example, if were to
//dispose of the list when the user clicked on Save, we may hit a snag.		It
//is possible for the user to specify the name of a file that already
//exists.	 When that happens, s/he is presented with another dialog that
//asks if they are sure of what they are doing.		At that point, the user
//could press Cancel, and return us to the PutFile dialog, SANS our second
//list! Neither can we defer disposing of list until Standard File is done
//and returns to our application, as the Dialog box we were using is gone,
//and the List Manager tries to perform an InvalRect on a non-existent
//window when it erases its list and that will crash.

//There are two solutions to this problem:

//* Implement the list as a Custom Control.		Then, when it comes time to
// close the dialog box, the custom control will get called with a dispCntl
// message.		It can take that opportunity to call LDispose.		This technique
// is not presented here, as it is more appropriate for a Custom Control
// Sample program.	However, it is the suggested way to proceed.

//* Ensure that the confirmation dialog box never comes up under Standard
// File's control.	By calling it ourself, we know what the user chooses.	 If
// the user presses OK, we delete the offending file, dispose of our list,
// and return to Standard File.		If the user presses Cancel, we change the
// click on the Save button into a NULL event.	This is the algorithm we use
// below.

//There is a major disadvantage with this last approach, however.	 With it,
//we cannot implement a safe saving procedure.	Normally, the best way to
//save a file (disk space permitting), is to save the data to a temporary
//file, delete the original, and then rename the temporary file to that of
//the original.		However, with the second approach, the file is deleted well
//before it is safe to do so.

//This example *could* be reworked to avoid this problem.	 For instance,
//instead of being deleted, the offending file could be renamed to
//something else or moved to another directory.

//DoPutListsFile
//--------------
//We force the current folder of Standard File to be the location of the
//temporary file that we've created.	This is to attempt to make the
//Existing File? dialog show up.	Being a user friendly application, we
//save and restore the user's current folder.	 If the user did happen to
//select to replace the file, we create a new one for the next time.

//PutListsDlglFilter
//------------------
//The Dialog Filter is used to handle clicks in the List Manager list.

//PutListsDlgHook
//---------------
//The initialization case is used to create a list of file types.	 It also
//jams in the name of the temporary file that we are using.		This is to try
//and force the "Replace file?" dialog to show up.	This is only part of
//the demonstration, not something that programs would really be doing.		If
//the user attempts to save the file with the name of an existing file,
//then the Existing File alert needs to be shown.	 This is centered within
//the main dialog alert properly according to the Human Interface Guidelines.

//LNPFDrawList
//------------
//Dialog Manager's user item drawing procedure to update the List Manager
//list.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void LNPFDrawList(WindowPtr theWindow, short item)
{
	if (item == kListItem) {
	LUpdate(theWindow->visRgn, gListHandle);
	DrawItemFrame(theWindow, kListItem);
	}
}

short PutListsDlgHook(short item, DialogPtr dlgPtr)
{
	Str63 name;
	FInfo fndrInfo;
	Point dLoc;
	Rect iRect;
	Rect alertRect;
	Handle iHandle;
	AlertTHndl alertHandle;
	long parID;
	short vRefNum;
	short replaceItem;
	short iKind;
	short choice;
	OSErr err;
	Str63 Temp_Str63_1;

	replaceItem = item;
	vRefNum = GetSFVRefNum();
	parID = GetSFCurDir();
	switch (replaceItem) { 
	
	case kFirstTime:
		GetDialogItem(dlgPtr, kListItem, &iKind, &iHandle, &iRect);
		CreateFPutList(dlgPtr, &iRect, &gListHandle, gPutFormat);
		SetDialogItem(dlgPtr, kListItem, iKind, (Handle)LNPFDrawList, 
					&iRect);
		memcpy(&name, GetTempFileName((Str63 *)Temp_Str63_1), sizeof(Str63));
		GetDialogItem(dlgPtr, putName, &iKind, &iHandle, &iRect);
		SetDialogItemText(iHandle, name);
		TESetSelect(0, PStr_Length(name), ((DialogPeek)dlgPtr)->textH);
		break;
	
	case putSave:
		err = HGetFInfo(vRefNum, parID, gReply.fName, &fndrInfo);
		if (err == noErr) {							//if GetFInfo = noErr, then
							//this file already exists.
							
		alertHandle = (AlertTHndl)Get1Resource('ALRT', kExistingFileALRT);
		HNoPurge((Handle)alertHandle);				//keep is around
		alertRect = (*alertHandle)->boundsRect;
		GlobalToLocal(&alertRect.topLeft);
		GlobalToLocal(&alertRect.botRight);
		PositionTwoRects(&dlgPtr->portRect, &alertRect, FixRatio(1, 
																	2), 
							FixRatio(1, 3));
		LocalToGlobal(&alertRect.topLeft);
		LocalToGlobal(&alertRect.botRight);
		(*alertHandle)->boundsRect = alertRect;
		ParamText(gReply.fName, "\p", "\p", "\p");
		choice = Alert(kExistingFileALRT, NULL);
		HPurge((Handle)alertHandle);				//let it go
		if (choice == kReplaceItem) 
			err = HDelete(vRefNum, parID, gReply.fName);
		else										//Change "Save" into null event
			replaceItem = kNullModalEvt;
		}
		break;
	}
	
	//CASE
	if (replaceItem == putSave || replaceItem == putCancel) 
	DisposeFPutList(gListHandle, gPutFormat);
	return replaceItem;
}

Boolean PutListsDlglFilter(DialogPtr dlgPtr, EventRecord *evt, SInt16 *itemHit)
{
	Point localPt;
	Boolean ignore;

	if (evt->what == mouseDown) {
	localPt = evt->where;
	GlobalToLocal(&localPt);
	ignore = LClick(localPt, evt->modifiers, gListHandle);
	}
	return false;
}

void DoPutListsFile(void)
{
	Str255 msg;										//prompt String
	Str63 fName;
	long oldDirStore;
	short oldVRefNum;
	short err;
	DlgHookUPP dhUPP;
	ModalFilterUPP mfUPP;
	Str63 Temp_Str63_1;

	oldDirStore = GetSFCurDir();						//save old folder and volume
	oldVRefNum = GetSFVRefNum();
	SetSFCurDir(gAppParID);							//force to our app folder
	SetSFVRefNum(gAppVRefNum);
	
	GetIndString(msg, rStrMisc, sSaveAsMsg);
	
	dhUPP = NewDlgHookProc((DlgHookProcPtr)PutListsDlgHook);
	mfUPP = NewModalFilterProc((ModalFilterProcPtr)PutListsDlglFilter);
	SFPPutFile((Point)kSFTopLeft,						//location
				msg,									//prompt String
				"\p",									//original name
				dhUPP,									//dialog hook
				&gReply,								//record for returned values
				rPutListsFileDLOG,						//ID of custom Dialog
				mfUPP);								//ModalDialog filterProc
	DisposeRoutineDescriptor(dhUPP);
	DisposeRoutineDescriptor(mfUPP);
	
	if (gReply.good) {
	ShowSelection(gReply.fName, gPutFormat);
	memcpy(&fName, GetTempFileName((Str63 *)Temp_Str63_1), sizeof(Str63));
	if (PStr_Cmp(gReply.fName, fName) == 0) 
		err = HCreate(gAppVRefNum, gAppParID, fName, rAppSignature, 
					'TEMP');
	} else 
	ShowCancelled();
	
	SetSFCurDir(oldDirStore);							//restore old folder and volume
	SetSFVRefNum(oldVRefNum);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//DoPutOptions with a Dialog Hook
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This demonstration add a new button to the SFPutFile dialog.	This button
//is used to bring up another dialog used to show a set of radio buttons.
//The options dialog allows the user to select a file type, which will be
//used to as the file format when saving the file.

//DoPutOptions
//------------
//Very simple version of creating a SFPPutFile dialog with a Dialog Hook
//that checks for the addition button being selected.

//PutOptionsDlgHook
//-----------------
//If the Options button is selected, then we bring up the options dialog.
//The default button is outlined and we wait for the OK or Cancel item.
//The new format is the item that was hit within the options dialog.	The
//title of this item is displayed to the user in the SFPutFile dialog
//showing the new file format.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

short PutOptionsDlgHook(short item, DialogPtr dlgPtr)
{
	Str255 title;
	DialogTHndl dlogTemplate;
	DialogPtr subDlgPtr;
	Rect dlogRect;
	Rect iRect;
	Handle iHandle;
	short itemHit;
	short iKind;
	short newItem;
	GrafPtr oldPort;
	short value_PutOptionsDlgHook = item;

	switch (item) { 
	
	case kOptionsButton:
		GetPort(&oldPort);
		
		dlogTemplate = (DialogTHndl)Get1Resource('DLOG', rOptionsSubDLOG);
		dlogRect = (*dlogTemplate)->boundsRect;
		GlobalToLocal(&dlogRect.topLeft);
		GlobalToLocal(&dlogRect.botRight);
		PositionTwoRects(&dlgPtr->portRect, &dlogRect, FixRatio(1, 2), 
						FixRatio(1, 3));
		LocalToGlobal(&dlogRect.topLeft);
		LocalToGlobal(&dlogRect.botRight);
		(*dlogTemplate)->boundsRect = dlogRect;
		
		subDlgPtr = GetNewDialog(rOptionsSubDLOG, NULL, (WindowPtr)(-1));
		GetDialogItem(subDlgPtr, ok, &iKind, &iHandle, &iRect);
		OutlineButton((ControlHandle)iHandle);
		SetRadioButton(subDlgPtr, kDefaultFormat, kCntlOn);
		GetDialogItem(subDlgPtr, kFrameItem, &iKind, &iHandle, &iRect);
		SetDialogItem(subDlgPtr, kFrameItem, iKind, (Handle)DrawItemFrame, 
					&iRect);
		newItem = kDefaultFormat;
		do {
		ModalDialog(NULL, &itemHit);
		if (itemHit != newItem && itemHit > cancel) {
			SetRadioButton(subDlgPtr, newItem, kCntlOff);
			SetRadioButton(subDlgPtr, itemHit, kCntlOn);
			newItem = itemHit;
		}
		} while (!(itemHit == ok || itemHit == cancel));
		
		if (itemHit == ok) {
		GetDialogItem(subDlgPtr, newItem, &iKind, &iHandle, &iRect);
		GetControlTitle((ControlHandle)iHandle, gPutFormat);
		GetDialogItem(dlgPtr, kFormatString, &iKind, &iHandle, &iRect);
		SetDialogItemText(iHandle, gPutFormat);
		}
		DisposeDialog(subDlgPtr);
		SetPort(oldPort);
		break;
	}
	
	//CASE
	return value_PutOptionsDlgHook;
}

void DoPutOptions(void)
{
	Str255 msg;										//prompt String
	SFReply reply;
	DlgHookUPP dhUPP;

	GetIndString(msg, rStrMisc, sSaveAsMsg);
	
	dhUPP = NewDlgHookProc((DlgHookProcPtr)PutOptionsDlgHook);
	SFPPutFile((Point)kSFTopLeft,						//location
				msg,									//prompt String
				"\pDoug",								//original name
				dhUPP,									//dialog hook
				&reply,								//record for returned values
				rOptionsDLOG,							//ID of custom Dialog
				NULL);									//ModalDialog filterProc
	DisposeRoutineDescriptor(dhUPP);
	
	if (reply.good) 
	ShowSelection(reply.fName, gPutFormat);
	else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//DoIdleUpdates
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This demonstrates two things: how to handle idle events and how to
//determine the currently selected file.	There is a problem that Standard
//File has when updates are pending in background windows of the current
//application partition.	Standard File calls ModalDialog at the heart of
//its main loop.	ModalDialog calls GetNextEvent, and then calls a
//filterProc internal to Standard File.		This filterProc performs some
//processing on the event, (like handling hits on the filename list, the
//Current Directory button, and the Disk Name Icon), and then calls the
//user filterProc specified in SFPPutFile or SFPGetFile calls.	Another one
//of the things that the internal filterProc does is look for a nullEvent.
//When one is found, the filterProc returns 100 as the "itemHit".	 This
//forces ModalDialog to return to Standard File, which in turn can now call
//the Dialog Hook with the manufactured item number 100, indicating that
//idle time processing can be performed.

//The problem occurs when there are updates pending in windows open in the
//current application's partition.	These updates will 'clog' the event
//queue.	When ModalDialog calls GetNextEvent, it will get the update event.
//However, there is no way for ModalDialog to respond to them.	It therefore
//passes the event off to the filterProc, which is Standard File's internal
//one.	This filterProc doesn't know how to handle it either.	 Normally, the
//event would be ignored at this point, and the update event would be
//unresolved.	 Control returns back to ModalDialog, which calls GetNextEvent
//again, and gets the another update event since it wasn't handled.		Null
//events will not get returned by GetNextEvent and, hence, the dialog hook
//will never get called with itemHit=100.

//This situation can be solved by providing your own filterProc procedure
//to be called after Standard File's's internal Dialog Filter.	It will be
//this routine's responsibility to check for update events, and handle them
//appropriately.	Usually, this would mean that the update procedure within
//the application would be called, and the update could be cleared by
//BeginUpdate/EndUpdate.	However, the filterProc could also just handle
//update events in the same way that Standard File's filterProc handles
//null events.	This is done by returning 100 in the ItemHit parameter and
//TRUE as the function result.	This is the approach taken by the sample
//below.

//DoIdleUpdates
//-------------
//Create a window with a non-empty update region then call up SFPGetFile
//with a Dialog Hook and Dialog Filter.

//IdleDlgFilter
//-------------
//If we get an update event for a window other than the dialog box, change
//it to a null event, and tell ModalDialog that we handled it.	This allows
//our Dialog Hook idle time.

//IdleDlgHook
//-----------
//All it does is wait around for null events and draws the current time
//when it gets one.		Note that this routine will NOT get called with
//item=100 if we did not have the IdleDlgFilter below.	Determining if
//there is an item selected is straight forward, once you know the secret.
//During the idle event the SFReply record will contain the file name if a
//file is selected.		(It's also worth noting here that if the fName field
//is NIL, this could mean a folder is selected.		If this were the case then
//fType will be the DirID of that folder!	 Pretty cool huh?)

//Given that we can determine the name of the file, it's easy to find out
//the size.		So, just for fun we'll show the file's size on the disk to the
//user.		ioFlPyLen is the physical size of the file on the device. Logical
//size would be the actual size of the file in memory.	In other words,
//GetEOF returns the logical size.	The data and resource fork's physical
//sizes are added together to give the true size of the entire file on the
//volume.


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

short IdleDlgHook(short item, DialogPtr dlgPtr)
{
	Str255 newStr;
	Str255 itemStr;
	Str255 conCatStr;
	HParamBlockRec hPBRec;
	long dateTime;
	long totalSize;									//size of file in k bytes
	Rect iRect;
	Handle iHandle;
	short iKind;
	OSErr err;
	short value_IdleDlgHook = item;
	Str255 Temp_Str255_1;

	if (item == kNullModalEvt) {
	GetDateTime(&dateTime);
	IUTimeString(dateTime, kWantSeconds, newStr);
	GetDialogItem(dlgPtr, kTimeItem, &iKind, &iHandle, &iRect);
	GetDialogItemText(iHandle, itemStr);
	if (PStr_Cmp(newStr, itemStr) != 0) 
		SetDialogItemText(iHandle, newStr);
	
	GetDialogItem(dlgPtr, kSizeItem, &iKind, &iHandle, &iRect);
	GetDialogItemText(iHandle, itemStr);
	if (PStr_Cmp(gReply.fName, "\p") != 0) {
		hPBRec.ioParam.ioCompletion = NULL;
		hPBRec.ioParam.ioNamePtr = (StringPtr)gReply.fName;
		hPBRec.ioParam.ioVRefNum = GetSFVRefNum();
		hPBRec.fileParam.ioFDirIndex = 0;
		hPBRec.fileParam.ioDirID = GetSFCurDir();
		err = PBHGetFInfo(&hPBRec, !kFSAsynch);
		if (err == noErr) {
		PStr_Set(newStr, gReply.fName);
		newStr[0] = (char)((short)newStr[0] + 1);
		newStr[PStr_Length(newStr)] = ' ';
		
		totalSize = (hPBRec.fileParam.ioFlPyLen + hPBRec.fileParam.ioFlRPyLen) / 1024;
		NumToString(totalSize, conCatStr);
		PStr_Set(newStr, PStr_Concat(Temp_Str255_1, newStr, conCatStr, 
										NULL));
		
		GetIndString(conCatStr, rStrMisc, sFileSize);
		PStr_Set(newStr, PStr_Concat(Temp_Str255_1, newStr, conCatStr, 
										NULL));
		
		if (PStr_Cmp(newStr, itemStr) != 0) 
			SetDialogItemText(iHandle, newStr);
		} else 
		SetDialogItemText(iHandle, "\p??");
	} else 
		SetDialogItemText(iHandle, "\p");
	}
	return value_IdleDlgHook;
}

Boolean IdleDlgFilter(DialogPtr dlgPtr, EventRecord *evt, SInt16 *itemHit)
{
	Boolean value_IdleDlgFilter = false;

	if (evt->what == updateEvt && (DialogPtr)evt->message != dlgPtr) {
	*itemHit = kNullModalEvt;
	value_IdleDlgFilter = true;
	}
	return value_IdleDlgFilter;
}

void DoIdleUpdates(void)
{
	SFTypeList typeList;
	WindowPtr window;
	DlgHookUPP dhUPP;
	ModalFilterUPP mfUPP;

	window = GetNewWindow(rUpdateWindow, NULL, (WindowPtr)-1);
	
	dhUPP = NewDlgHookProc((DlgHookProcPtr)IdleDlgHook);
	mfUPP = NewModalFilterProc((ModalFilterProcPtr)IdleDlgFilter);
	SFPGetFile((Point)kSFTopLeft,						//location
				"\p",									//vestigial String
				NULL,									//file filter
				kShowAllFiles,							//numtypes
				typeList,								//array to types to show
				dhUPP,									//Dialog Hook
				&gReply,								//record for returned values
				rGetIdleUpdates,						//ID of Normal Dialog
				mfUPP);								//ModalDialog filterProc
	DisposeRoutineDescriptor(dhUPP);
	DisposeRoutineDescriptor(mfUPP);
	
	CloseWindow(window);
	if (gReply.good) 
	ShowSelection(gReply.fName, "\p");
	else 
	ShowCancelled();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//DoRememberFile
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//This demonstrates how to ask the user to locate a file and then how the
//application should remember it.	 The proper way to do this is use the
//volume name, directory ID and file name.	Do not use the vRefNum.		It
//could be a working directory which will change, probably as soon as your
//application quits.	It could also be a real vRefNum but that could change
//as soon as the user re-boots.		Directory ID do not change unless the user
//deletes the folder.	 No matter where the user places that folder or if
//they rename it, the DirID will be the same.

//Real applications should have a preference file.	This would store the
//users options, such as the location of a file.	For the purposes of this
//demonstration, we're using the application's resource fork to hold the
//file's location.	Some potential problem with the method is if the file
//is locked, the volume is lock, or the application is on a shared volume
//by many different users.

//RecallFile
//----------
//Get the application defined resource and see if the file it describes can
//be used to locate a file.		Start by changing the volume name into a
//vRefNum.	If this works, then a volume with that name is mounted.		Then
//we attempt to show a dialog that displays some information to the user
//proving we've located their file.		If we could not locate the file, then
//we ask them to locate another one and we'll try again next time.

//DoRememberFile
//--------------
//Call upon SFGetFile to ask the user to select a file.		Then change the
//working directory number returned in vRefNum into a real vRefNum and
//DirID.	The real vRefNum is then used to option the volume name.	This
//information along with the file name is saved into an application defined
//resource and saved.

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DoRememberFile(void)
{
	Str27 volName;
	SFTypeList typeList;
	SFReply reply;
	WDPBRec myWDPBRec;
	FileDescHdl rememberFile;
	OSErr err;

	SFGetFile((Point)kSFTopLeft,						//location
			"\p",									//prompt String
			NULL,									//file filter
			kShowAllFiles,							//numtypes
			typeList,								//array to types to show
			NULL,									//dialog hook
			&reply);								//record for returned values
	if (reply.good) {
	ShowSelection(reply.fName, "\p");
	PStr_Set(volName, "\p");						//best to empty volName
													//set up WDPBRec
	myWDPBRec.ioNamePtr = (StringPtr)volName;
	myWDPBRec.ioVRefNum = reply.vRefNum;
	myWDPBRec.ioWDIndex = 0;
	myWDPBRec.ioWDProcID = 0;
	err = PBGetWDInfo(&myWDPBRec, !kFSAsynch);
	if (err == noErr) {
		rememberFile = (FileDescHdl)Get1Resource('FILE', rMemorizedFile);
		if (rememberFile != NULL) {
		PStr_Set((*rememberFile)->name, reply.fName);
		(*rememberFile)->parID = myWDPBRec.ioWDDirID;
		PStr_Set((*rememberFile)->volName, volName);
		ChangedResource((Handle)rememberFile);
		WriteResource((Handle)rememberFile);
		} else 
		AlertUser(ResError(), sResErr);
	} else 
		AlertUser(err, sFileSystem);
	} else 
	ShowCancelled();
}

void RecallFile(void)
{
	FileDescHdl rememberFile;
	short vRefNum;
	OSErr err;

	rememberFile = (FileDescHdl)Get1Resource('FILE', rMemorizedFile);
	if (rememberFile != NULL) {
	HLock((Handle)rememberFile);
	FileDescPtr with_4 = *rememberFile;
	err = GetVol((StringPtr)with_4->volName, &vRefNum);
	if (err == noErr) 
		err = ShowFoundFile(with_4->name, with_4->parID, vRefNum);
	HUnlock((Handle)rememberFile);
	if (err != noErr) {								//catch any file errors
		AlertUser(noErr, sLostFile);
		DoRememberFile();
	}
	} else 
	AlertUser(ResError(), sResErr);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void DoCommand(long mResult)
{
	//Performed the selected menu command
	
	Str255 name;
	GrafPtr oldPort;
	short theItem;
	short theMenu;
	short daRefNum;
	Boolean ignore;

	theItem = LoWord(mResult);
	theMenu = HiWord(mResult);
	
	switch (theMenu) { 
	
	case mApple:
		if (theItem == iAboutMe) 
		DoAbout();
		else {
		GetMenuItemText(GetMenuHandle(mApple), theItem, name);
		GetPort(&oldPort);
		daRefNum = OpenDeskAcc(name);
		SetPort(oldPort);
		}
		break;
	
	case mFile:
		switch (theItem) { 
		case iQuit:
			Terminate();
			break;
		}
		break;
	
	case mEdit:
		ignore = SystemEdit(theItem - 1);
		break;
	
	case mSFExamples:
		switch (theItem) { 
		case iNormalGet:
			ShowPath();
			break;
		case iNormalPGet:
			DoNormalPGet();
			break;
		case iFileFilter:
			DoFFilter();
			break;
		case iGetDirectory:
			DoGetDirectory();
			break;
		case iMultiFile:
			DoMultiFile();
			break;
		case iNormalPut:
			DoNormalPut();
			break;
		case iNormalPPut:
			DoNormalPPut();
			break;
		case iForceDirectory:
			DoForceDirectory();
			break;
		case iPutListsFile:
			DoPutListsFile();
			break;
		case iPutOptions:
			DoPutOptions();
			break;
		case iIdleUpdates:
			DoIdleUpdates();
			break;
		case iRememberFile:
			DoRememberFile();
			break;
		} //CASE
		break;
	}
	HiliteMenu(0);									//unhighlight all menus
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void EventLoop(void)
{
	//Chances are, you've seen an event loop before so I'm not going to
	//comment about it.
	
	const char	osEvt = app4Evt;						//defined in post MPW 3.0 Events.p
	const char	suspendResumeMessage = 1;				//defined in post MPW 3.0 Events.p
	const char	kGetHighByte = 24;						//24 bits (three bytes) to shift right
	const char	kResumeMask = 1;						//resume event is low bit of message
	
	EventRecord evt;
	WindowPtr whichWindow;							//used for FindWindow
	short deskPart;									//result from FindWindow
	OSErr err;
	Boolean haveEvent;								//TRUE if interesting event

	do {
	if (gHasWaitNextEvent) 
		haveEvent = WaitNextEvent(everyEvent, &evt, LONG_MAX, NULL);
	else {
		SystemTask();
		haveEvent = GetNextEvent(everyEvent, &evt);
	}
	
	deskPart = FindWindow(evt.where, &whichWindow);
	if (deskPart != inSysWindow) 
		SetCursor(&qd.arrow);
	
	if (haveEvent) {
		switch (evt.what) { 
		
		case mouseDown:
			switch (deskPart) { 
			case inSysWindow:
				SystemClick(&evt, whichWindow);
				break;
			case inMenuBar:
				DoCommand(MenuSelect(evt.where));
				break;
			}
			break;
		
		case keyDown:
		case autoKey:
			if (BAnd(evt.modifiers, cmdKey) != 0) 
			DoCommand(MenuKey((char)BAnd(evt.message, charCodeMask)));
			break;
		
		case diskEvt:
			//Call DIBadMount in response to a diskEvt
			if (HiWord(evt.message) != noErr) 
			err = DIBadMount((Point)kSFTopLeft, evt.message);
			break;
		
		case osEvt:
			switch (BSR(evt.message, kGetHighByte)) { 
			case suspendResumeMessage:
				if (BAnd(evt.message, kResumeMask) != 0) {
				gInBackground = false;
				} else 
				gInBackground = true;				//suspend event
				break;								//suspendResumeMessage
			}
			break;
		}
		
		//CASE
	}
	} while (true);									//loop forever, quit through Terminate
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Initialize(void)
{
	//Due to some bug in older versions of A/UX regarding working directories,
	//I check which verison we might be running if any.	 If A/UX isn't running,
	//then the version will be zero.	Other wise it will be set to the major
	//version number, such as 2 for 2.0.	I want to make sure that enough
	//memory is free for my application to run.	 It is possible that user may
	//have adjusted the SIZE resource to too small a setting or for some other
	//reason the application started up in a very small memory partition.		I
	//preform this check after initializing all the Toolbox and the basic
	//features of this application.	 The pentultimate thing is to save the
	//default directory and volume that the application is launched from in two
	//globals and create a temporary file.	This file is only used for
	//demonstration purposes.		Finally, we check if the file we last remembered
	//can be found.
	
	Str63			fName;
	EventRecord event;
	Handle		menuBar;
	Handle		apParam;
	OSErr			ignoreError;
	short			count;
	short			auxVersion;
	short			appResRef;
	OSErr			err;
	Boolean		ignoreResult;
	Str63			Temp_Str63_1;

	gInBackground = false;							//well be in the foreground soon
	gHasWaitNextEvent = TrapExists(_WaitNextEvent);
	
	for (count = 1; count <= kNumberOfMasters; ++count) //allocate master pointer blocks
	MoreMasters();
	InitGraf((Ptr)&qd.thePort);						//init managers, yawn...
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(NULL);
	InitCursor();
	
	//This code is necessary to pull the application into the foreground.		I use
	//EventAvail because I dont want to remove any events the user may have
	//done, such as typing ahead.		Until the application has made a few calls (3
	//seems to be the magic number) to the Event Manager, MultiFinder keeps me
	//in the background.		Splashscreens and Alerts will remain in a background
	//layer until we get a few events.	This is documented in Tech Note #180.
	
	for (count = 1; count <= kBroughtToFront; ++count) 
	ignoreResult = EventAvail(everyEvent, &event);
	
	auxVersion = GetAUXVersion();
	if (auxVersion < kMinAUXVersion && auxVersion > 0) 
	EmergencyExit(sOldAUX);
	
	ignoreError = SysEnvirons(kSysEnvironsVersion, &gMac);
	if (gMac.machineType < envMachUnknown)			//we use HFS
	EmergencyExit(sNoHFS);
	
	menuBar = GetNewMBar(rMenuBar);					//read menus into menu bar
	if (menuBar == NULL) 
	EmergencyExit(sNoMenus);
	SetMenuBar(menuBar);								//install menus
	DisposeHandle(menuBar);
	AppendResMenu(GetMenuHandle(mApple), 'DRVR');		//add DA names to Apple menu
	DrawMenuBar();
	
	if (FailLowMemory(kMinSpace)) 
	EmergencyExit(sLowMemory);
	
	GetAppParms(gAppName, &appResRef, &apParam);
	err = MyHGetVol(NULL, &gAppVRefNum, &gAppParID);
	err = HCreate(gAppVRefNum, gAppParID, GetTempFileName((Str63 *)Temp_Str63_1), 
				rAppSignature, 'TEMP');
	
	RecallFile();										//what was that file?
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ MAIN PROGRAM ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void main()
{	 
	//If you have stack requirements that differ from the default, then you
	//could use SetApplLimit to increase StackSpace at this point, before
	//calling MaxApplZone.
	
	MaxApplZone();				//expand the heap so code segments load at the top
	Initialize();				//initialize the program
	UnloadSeg(Initialize);		//note: Initialize must not be in Main!
	EventLoop();				//call the main event loop
}
